==> web-ui-pack-0.4.1.tgz: package/helpers/observer.js <==
/* eslint-disable prefer-rest-params */
// discussions here https://stackoverflow.com/questions/5100376/how-to-watch-for-array-changes
import isEqual from "./isEqual";
// #region Helpers
const arrRemove = (arr, item) => {
    const i = arr.indexOf(item);
    /* istanbul ignore else */
    if (i > -1) {
        arr.splice(i, 1);
    }
};
const some = (m, predicate) => {
    // eslint-disable-next-line no-restricted-syntax
    for (const [key, value] of m.entries()) {
        /* istanbul ignore else */
        if (predicate(key, value)) {
            return true;
        }
    }
    return false;
};
const pathThrough = (fn) => new Promise((resolve) => {
    resolve(fn());
});
const isObject = (obj) => typeof obj === "object" && obj !== null && !(obj instanceof HTMLElement);
const isDate = (obj) => Object.prototype.toString.call(obj) === "[object Date]";
// #endregion
// #region Functions
const watchSet = [
    {
        is: isDate,
        keys: new Set([
            "setDate",
            "setFullYear",
            "setHours",
            "setMilliseconds",
            "setMinutes",
            "setMonth",
            "setSeconds",
            "setTime",
            "setUTCDate",
            "setUTCFullYear",
            "setUTCHours",
            "setUTCMilliseconds",
            "setUTCMinutes",
            "setUTCMonth",
            "setUTCSeconds",
            "setYear",
        ]),
        getVal: (obj) => obj.valueOf(),
        propKey: "valueOf",
    },
    {
        is: (obj) => Object.prototype.toString.call(obj) === "[object Set]",
        keys: new Set(["add", "delete", "clear"]),
        getVal: (obj) => obj.size,
        propKey: "size",
    },
    {
        is: (obj) => Object.prototype.toString.call(obj) === "[object Map]",
        keys: new Set(["set", "delete", "clear"]),
        getVal: (obj) => obj.size,
        propKey: "size",
    },
];
/** list with pair <Proxy, Ref>; Ref is bunch of arrays, callbacks etc. */
const lstObserved = new WeakMap();
/** list with pair <RawObject, Proxy>; */
const lstObjProxy = new WeakMap();
function isObserved(obj) {
    return obj && lstObserved.has(obj);
}
function appenCallback(proxy, callback, setKey) {
    const o = lstObserved.get(proxy);
    if (!o) {
        throw new Error("Observer. Only observed objects expected. Use make() before");
    }
    if (!callback) {
        throw new Error("Observer. Callback is missed");
    }
    const listeners = o[setKey];
    listeners.push(callback);
    return () => arrRemove(listeners, callback);
}
function make(obj, parentRef) {
    // checking if object is already observed
    const prevProxy = lstObjProxy.get(obj);
    const isAssigned = lstObserved.get(prevProxy || obj);
    if (isAssigned) {
        if (parentRef) {
            // if (isAssigned.parentRefs.has(parentRef.ref)) {
            //   throw new Error("Observer. Wrong assignment: parentRef must be unique");
            // }
            isAssigned.parentRefs.set(parentRef.ref, parentRef.key);
        }
        return prevProxy || obj;
    }
    const isDateObj = isDate(obj);
    const ref = {
        propListeners: [],
        listeners: [],
        parentRefs: parentRef ? new Map([[parentRef.ref, parentRef.key]]) : new Map(),
        hasListeners: () => !!ref.listeners.length || //
            !!ref.propListeners.length || //
            some(ref.parentRefs, (r) => r.hasListeners()),
        hasObjListeners: () => !!ref.listeners.length || some(ref.parentRefs, (r) => r.hasObjListeners()),
        onPropChanged: (e) => {
            const target = proxy;
            if (e.target === null) {
                e.prev = target[e.prop];
                e.next = target[e.prop];
            }
            e.target = target; // required to reassign and propagate event through parents
            ref.propListeners.forEach((f) => pathThrough(() => f(e)));
            ref.parentRefs.forEach((key, r) => r.onPropChanged({ ...e, prop: key, target: null }));
        },
        onChanged: (e) => {
            e.target = proxy; // required to reassign and propagate event through parents
            ref.listeners.forEach((f) => pathThrough(() => f(e)));
            ref.parentRefs.forEach((key, r) => r.onChanged({ ...e, props: [key] }));
        },
    };
    let changedProps = [];
    let timeoutId;
    const propChanged = (e) => {
        e.target = proxy;
        ref.onPropChanged(e);
        if (ref.hasObjListeners()) {
            changedProps.push(e.prop);
            if (timeoutId) {
                return;
            }
            // empty timeout filters bunch of changes to single event
            timeoutId = setTimeout(() => {
                timeoutId = undefined;
                const ev = {
                    props: changedProps,
                    target: proxy,
                };
                changedProps = [];
                ref.onChanged(ev);
            });
        }
    };
    let isReady = false;
    const proxyHandler = {
        set(t, prop, next, receiver) {
            if (!isReady) {
                return Reflect.set(t, prop, next, receiver);
            }
            const prev = t[prop];
            if (prev !== next) {
                // prev can be Proxy or Raw
                // next can be Proxy or Raw
                lstObserved.get(prev)?.parentRefs.delete(ref);
                if (isObject(next)) {
                    next = make(next, { key: prop, ref });
                }
            }
            const isOk = Reflect.set(t, prop, next, receiver);
            if (isOk && ref.hasListeners()) {
                !isEqual(prev, next) && propChanged({ prev, next, prop });
            }
            return isOk;
        },
        deleteProperty(t, prop) {
            const prev = t[prop];
            debugger;
            if (!isDateObj && ref.hasListeners()) {
                propChanged({ prev, next: undefined, prop });
            }
            // remove parent from this object
            if (isObject(prev)) {
                const v = proxy[prop];
                lstObserved.get(v).parentRefs.delete(ref);
            }
            return true;
        },
    };
    const watchObj = watchSet.find((v) => v.is(obj));
    if (watchObj) {
        proxyHandler.get = function get(t, prop, receiver) {
            const v = Reflect.get(t, prop, receiver);
            // wrap if listeners exists
            if (typeof v === "function") {
                if (watchObj.keys.has(prop) && ref.hasListeners()) {
                    return (...args) => {
                        const prev = watchObj.getVal(t);
                        const r = v.apply(t, args);
                        const next = watchObj.getVal(t);
                        // Date possible to be NaN;
                        !isEqual(prev, next) && propChanged({ prev, next, prop: watchObj.propKey });
                        return r;
                    };
                }
                // return default with correct context
                return v.bind(t);
            }
            return v;
        };
    }
    const proxy = new Proxy(obj, proxyHandler);
    lstObserved.set(proxy, ref);
    lstObjProxy.set(obj, proxy);
    if (isObject(obj) && obj.valueOf() === obj) {
        // warn: possible error if object isn't extensible
        Object.defineProperty(proxy, "valueOf", { value: () => obj.valueOf, enumerable: false });
    }
    // scan recursive
    // it doesn't required because object keys is null: if (!isDate && !(obj instanceof Map || obj instanceof Set)) {
    Object.keys(obj).forEach((k) => {
        const v = obj[k];
        if (isObject(v)) {
            proxy[k] = make(v, { ref, key: k });
        }
    });
    // }
    isReady = true; // required to avoid double make() on same proxy
    return proxy;
}
// #endregion
/** Helper to detect object changes;
 * @example
 * const raw = { date: new Date(), period: 3, nestedObj: { val: 1} };
 * const obj = observer.make(raw);
 * const removeListener = observer.onPropChanged(obj, (e) => console.warn(e));
 * const removeListener2 = observer.onChanged(obj, (e) => console.warn(e));
 * obj.period = 5; // events are fired after 1ms
 * removeListener();
 * removeListener2();
 */
const observer = {
    isObserved,
    make(obj) {
        return make(obj, undefined);
    },
    onPropChanged(target, callback
    // options: { once?: boolean }
    ) {
        return appenCallback(target, callback, "propListeners");
    },
    onChanged(target, callback
    // options: { once?: boolean }
    ) {
        return appenCallback(target, callback, "listeners");
    },
};
Object.seal(observer);
export default observer;
/*
  troubleshooting:
  const obj = observer.make({v: 1})
  obj.v = 2; // propsChanged
  obj.v = 1 // props not changed but event is called (because we use async)
*/

==> web-ui-pack-0.8.1.tgz: package/helpers/observer.js <==
/* eslint-disable prefer-rest-params */
// discussions here https://stackoverflow.com/questions/5100376/how-to-watch-for-array-changes
import isEqual from "./isEqual";
// #region Helpers
const arrRemove = (arr, item) => {
    const i = arr.indexOf(item);
    arr.splice(i, 1);
};
const some = (m, predicate) => {
    // eslint-disable-next-line no-restricted-syntax
    for (const [key, value] of m.entries()) {
        if (predicate(key, value)) {
            return true;
        }
    }
    return false;
};
const pathThrough = (fn) => new Promise((resolve) => {
    resolve(fn());
});
/** Returs true if object is valid Record<string, any> and need to iterate props */
const isRecord = (obj) => obj !== null && typeof obj === "object" && !(obj instanceof HTMLElement) && !(obj instanceof Promise);
const isDate = (obj) => Object.prototype.toString.call(obj) === "[object Date]";
// #endregion
// #region Functions
const watchSet = [
    {
        is: isDate,
        keys: new Set([
            "setDate",
            "setFullYear",
            "setHours",
            "setMilliseconds",
            "setMinutes",
            "setMonth",
            "setSeconds",
            "setTime",
            "setUTCDate",
            "setUTCFullYear",
            "setUTCHours",
            "setUTCMilliseconds",
            "setUTCMinutes",
            "setUTCMonth",
            "setUTCSeconds",
            "setYear",
        ]),
        getVal: (obj) => obj.valueOf(),
        propKey: "valueOf",
    },
    {
        is: (obj) => Object.prototype.toString.call(obj) === "[object Set]",
        keys: new Set(["add", "delete", "clear"]),
        getVal: (obj) => obj.size,
        propKey: "size",
    },
    {
        is: (obj) => Object.prototype.toString.call(obj) === "[object Map]",
        keys: new Set(["set", "delete", "clear"]),
        getVal: (obj) => obj.size,
        propKey: "size",
    },
];
/** list with pair <Proxy, Ref>; Ref is bunch of arrays, callbacks etc. */
const lstObserved = new WeakMap();
/** list with pair <RawObject, Proxy>; */
const lstObjProxy = new WeakMap();
function isObserved(obj) {
    return obj && lstObserved.has(obj);
}
function appendCallback(proxy, callback, setKey) {
    const o = lstObserved.get(proxy);
    if (!o) {
        throw new Error("Observer. Only observed objects expected. Use make() before");
    }
    if (!callback) {
        throw new Error("Observer. Callback is missed");
    }
    const listeners = o[setKey];
    listeners.push(callback);
    return () => arrRemove(listeners, callback);
}
function make(obj, parentRef, opts) {
    // checking if object is already observed
    const prevProxy = lstObjProxy.get(obj);
    const isAssigned = lstObserved.get(prevProxy || obj);
    if (isAssigned) {
        if (parentRef) {
            // if (isAssigned.parentRefs.has(parentRef.ref)) {
            //   throw new Error("Observer. Wrong assignment: parentRef must be unique");
            // }
            isAssigned.parentRefs.set(parentRef.ref, parentRef.key);
        }
        return prevProxy || obj;
    }
    const isDateObj = isDate(obj);
    const ref = {
        propListeners: [],
        listeners: [],
        parentRefs: parentRef ? new Map([[parentRef.ref, parentRef.key]]) : new Map(),
        hasListeners: () => !!ref.listeners.length || //
            !!ref.propListeners.length || //
            some(ref.parentRefs, (r) => r.hasListeners()),
        hasObjListeners: () => !!ref.listeners.length || some(ref.parentRefs, (r) => r.hasObjListeners()),
        onPropChanged: (e) => {
            const target = proxy;
            if (e.target === null) {
                e.prev = target[e.prop];
                e.next = target[e.prop];
            }
            e.target = target; // required to reassign and propagate event through parents
            ref.propListeners.forEach((f) => pathThrough(() => f(e)));
            ref.parentRefs.forEach((key, r) => r.onPropChanged({ ...e, prop: key, target: null }));
        },
        onChanged: (e) => {
            e.target = proxy; // required to reassign and propagate event through parents
            ref.listeners.forEach((f) => pathThrough(() => f(e)));
            ref.parentRefs.forEach((key, r) => r.onChanged({ ...e, props: [key] }));
        },
    };
    let changedProps = [];
    let timeoutId;
    const propChanged = (e) => {
        e.target = proxy;
        ref.onPropChanged(e);
        if (ref.hasObjListeners()) {
            changedProps.push(e.prop);
            if (timeoutId) {
                return;
            }
            // empty timeout filters bunch of changes to single event
            timeoutId = setTimeout(() => {
                timeoutId = undefined;
                const ev = {
                    props: changedProps,
                    target: proxy,
                };
                changedProps = [];
                ref.onChanged(ev);
            });
        }
    };
    let isReady = false;
    const proxyHandler = {
        set(t, prop, next, receiver) {
            if (!isReady) {
                return Reflect.set(t, prop, next, receiver);
            }
            const prev = t[prop];
            if (prev !== next) {
                // prev can be Proxy or Raw
                // next can be Proxy or Raw
                lstObserved.get(prev)?.parentRefs.delete(ref);
                if (isRecord(next)) {
                    const exclude = opts?.excludeNested;
                    const optsDeep = Array.isArray(exclude) && exclude.includes(prop) ? { ...opts, excludeNested: true } : opts;
                    next = make(next, { key: prop, ref }, optsDeep);
                }
            }
            const isOk = Reflect.set(t, prop, next, receiver);
            if (isOk && ref.hasListeners()) {
                !isEqual(prev, next) && propChanged({ prev, next, prop });
            }
            return isOk;
        },
        deleteProperty(t, prop) {
            const prev = t[prop];
            debugger;
            if (!isDateObj && ref.hasListeners()) {
                propChanged({ prev, next: undefined, prop });
            }
            // remove parent from this object
            if (isRecord(prev)) {
                const v = proxy[prop];
                lstObserved.get(v).parentRefs.delete(ref);
            }
            return true;
        },
    };
    const watchObj = watchSet.find((v) => v.is(obj));
    if (watchObj) {
        proxyHandler.get = function get(t, prop, receiver) {
            const v = Reflect.get(t, prop, receiver);
            // wrap if listeners exists
            if (typeof v === "function") {
                if (watchObj.keys.has(prop) && ref.hasListeners()) {
                    return (...args) => {
                        const prev = watchObj.getVal(t);
                        const r = v.apply(t, args);
                        const next = watchObj.getVal(t);
                        // Date possible to be NaN;
                        !isEqual(prev, next) && propChanged({ prev, next, prop: watchObj.propKey });
                        return r;
                    };
                }
                // return default with correct context
                return v.bind(t);
            }
            return v;
        };
    }
    const proxy = new Proxy(obj, proxyHandler);
    lstObserved.set(proxy, ref);
    lstObjProxy.set(obj, proxy);
    if (isRecord(obj) && obj.valueOf() === obj) {
        // warn: possible error if object isn't extensible
        Object.defineProperty(proxy, "valueOf", { value: () => obj.valueOf, enumerable: false });
    }
    // scan recursive
    const exclude = opts?.excludeNested;
    if (exclude !== true) {
        // not required because object keys is null: if (!isDate && !(obj instanceof Map || obj instanceof Set)) {
        Object.keys(obj).forEach((k) => {
            const v = obj[k];
            if (isRecord(v)) {
                const optsDeep = exclude && exclude.includes(k) ? { ...opts, excludeNested: true } : opts;
                proxy[k] = make(v, { ref, key: k }, optsDeep);
            }
        });
    }
    // }
    isReady = true; // required to avoid double make() on same proxy
    return proxy;
}
// #endregion
/** Helper to detect object changes;
 * @example
 * const raw = { date: new Date(), period: 3, nestedObj: { val: 1} };
 * const obj = observer.make(raw);
 * const removeListener = observer.onPropChanged(obj, (e) => console.warn(e));
 * const removeListener2 = observer.onChanged(obj, (e) => console.warn(e));
 * obj.period = 5; // events are fired after 1ms
 * removeListener();
 * removeListener2(); */
const observer = {
    isObserved,
    make(obj, opts) {
        return make(obj, undefined, opts);
    },
    onPropChanged(target, callback
    // options: { once?: boolean }
    ) {
        return appendCallback(target, callback, "propListeners");
    },
    onChanged(target, callback
    // options: { once?: boolean }
    ) {
        return appendCallback(target, callback, "listeners");
    },
};
Object.seal(observer);
export default observer;
/*
  troubleshooting:
  const obj = observer.make({v: 1})
  obj.v = 2; // propsChanged
  obj.v = 1 // props not changed but event is called (because we use async)
*/
